/*
 * bridged.c
 *
 * Erwin Kalvelagen <kalvelagen@fel.tno.nl>
 *
 * C version of Java version of bridge
 *
 * This fixes a problem with Illustra. When connecting with
 * Illustra from a remote client the socket connection may
 * deadlock. As a workaround the remote client now talks to
 * bridged at port 7532 which just passes on all info to
 * Illustra's midaemon/miserver at port 7531:
 *
 *
 *           
 *  client   ---------> server:7532  -----------> server:7531
 *           <--------               <-----------
 *           compressed
 *
 *
 * We also compression compression on traffic from the
 * server to the client (the other way around are likely
 * to be small queries, so no compression is used there 
 * -- i.e. inserts will be uncompressed!). On the server
 * side we used zlib 1.0.4 for this and on the client
 * the java.util.zip package.
 * 
 * Installs as a standalone running daemon or can be used
 * via inetd.
 *
 * Usage:
 *     bridged         when to run as a daemon
 *     bridged inetd   when running under inetd
 *
 * When installing under inetd, add a line to /etc/services:
 *   bridged 7532/tcp # MIJDBC/Illustra bridge
 * and a line to /etc/inetd.conf:
 *   bridged stream tcp nowait someuser /usr/local/bin/bridged bridged inetd
 *
 * When running as a daemon, a proper daemon process is created
 * and the parent exits. The daemon listens to port 7532 for incoming
 * connections. As soon as a client is asking for attention, a new
 * process is forked off that is set up just as inetd would do.
 * I.e. client socket is on stdin/stdout. This new server process
 * deals with the client while the parent continues to listen
 * to incoming connections.
 *
 * Change log
 * EK, August 1 1997, Initial Implementation
 * EK, August 21 1997. Sometimes both parent and child die.
 *                  Don't know why, but I suspect a SIGPIPE.
 * EK, Sept. 8, changed working dir to /tmp so a core can be written
 * EK, Sept. 22, close fd in parent (was possibly running out of fd's)
 *               child now uses _exit instead of exit
 *               catch SIGCHLD so no more defunct processes
 *               tested with 3 stresstest apps running concurrently 
 */


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <netinet/in.h>  
#include <errno.h>
#include <time.h>
#include <signal.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define ILLUSTRA_PORT 7531  /* port listened to by midaemon/miserver */
#define BRIDGE_PORT 7532    /* port we are listening to */

#define BUFSIZE 8192

#define LOGFILE "/tmp/bridged.log"

#define VERSION "bridged Version 0.1"
#define AUTHOR  "Erwin Kalvelagen, TNO-FEL, <kalvelagen@fel.tno.nl>"

#define DEBUG TRUE
#define COMPRESS


int init_daemon(void);
int run_daemon(void);
int run_server(void);
int handle_connection(int fd);
int logerror(char *s);
int logmessage(char *s);
void catch_signal(int sig);
void catch_sigchld(int sig);




#ifdef COMPRESS
int compress (unsigned char *dest, unsigned long *destLen, 
	      const unsigned char *source, unsigned long sourceLen);
#define BUFSIZEC (BUFSIZE+100)
/* actually enough is BUFSIZE*1.001+8 */
int logstatistics(int request_bytes, int answer_bytes, int compressed_answer_bytes);
#else
int logstatistics(int request_bytes, int answer_bytes);
#endif

int server_sock = -1; /* socket the deamon is listening to */
int client_sock = -1; /* socket connection to Illustra */




/*--------------------------------------------------------
 * main
 * depending on presence of command line argument 'inetd'
 * we assume stdin/stdout are our remote client socket or
 * install ourselves as a daemon
 *--------------------------------------------------------
 */
int main(int argc, char* argv[])
{

  int rc;
  int run_as_daemon = TRUE;


  if (argc > 1) 
    if (strcmp(argv[1],"inetd") == 0) 
      run_as_daemon = FALSE;


  if (run_as_daemon) {

    if ((rc = init_daemon()) < 0)
      return(rc);

    return (run_daemon());
  }
 
  else 
    run_server();
}


/*--------------------------------------------------------
 * init_daemon
 * initialize daemon, basically boilerplate code
 *--------------------------------------------------------
 */
int init_daemon(void)
{
  pid_t pid;
  int fd;

  if (DEBUG)
    logmessage("init_daemon");

  fprintf(stderr,"%s\n",VERSION);
  fprintf(stderr,"%s\n",AUTHOR);


  /*
   * ignore these signals
   */
  signal (SIGTTOU, SIG_IGN);
  signal (SIGTTIN, SIG_IGN);
  signal (SIGTSTP, SIG_IGN);


  if ((pid = fork()) < 0) 
    return(-1);
  else if (pid != 0) 
    exit(0); /* parent goes bye-bye */

  /* child continues */
  setsid();   /* become session leader */
  for (fd = 0; fd < NOFILE; ++fd)  /* close all (possible) open files */
    close(fd);
  chdir("/tmp"); /* so my workingdir doesn't get unmounted */
  umask(0);
  return(0);
  
}



/*--------------------------------------------------------
 * run_daemon
 * wait for incoming connections
 *--------------------------------------------------------
 */

int run_daemon(void) 
{
  int fd;
  struct sockaddr_in server_name;
  struct sockaddr_in client_name;
  size_t len;
  sigset_t mask;

  logmessage("run_daemon");


  /*
   * catch these signals 
   */
  signal (SIGTERM, catch_signal);
  signal (SIGPIPE, catch_signal); 



  /* 
   * Create a socket
   */
  server_sock = socket(PF_INET, SOCK_STREAM, 0);
  if (server_sock == -1) {
    logerror("socket:");
    return(-1);
  }

  
  /*
   * Give the socket a name. 
   */
  server_name.sin_family = AF_INET;
  server_name.sin_port = htons (BRIDGE_PORT);  /* our port! */
  server_name.sin_addr.s_addr = htonl (INADDR_ANY); /* means localhost */
  if (bind (server_sock, (struct sockaddr *) &server_name, sizeof (server_name)) < 0) {
      logerror ("bind");
      return(-1);
    }
 
  /*
   * listen for incoming requests
   */

  if (listen(server_sock, 4) < 0) {
    logerror("listen");
    return(-1);
  }
  

  /*
   * wait until an incoming request is accepted
   * and deal with that (in a separate proces)
   */


  sigemptyset(&mask);
  sigaddset(&mask,SIGCHLD);
  sigaddset(&mask,SIGHUP);
  sigaddset(&mask,SIGALRM);
    
  for(;;) {

    /*
     * block signals
     */
    if (sigprocmask(SIG_BLOCK,&mask,NULL)<0)
      logerror("sigprocmask");

    len = sizeof(client_name);
    fd = accept(server_sock, (struct sockaddr *)&client_name, &len); 
    if (fd < 0) {
      logerror("accept");
      return(-1);
    }

 
    /*
     * unblock signals and catch SIGCHLD
     */
    if (sigprocmask(SIG_UNBLOCK,&mask,NULL)<0)
      logerror("sigprocmask");
    signal (SIGCHLD, catch_sigchld);

   
    handle_connection(fd); 

    /*
     * can I close fd?
     */

    close(fd); 

  }



}


/*--------------------------------------------------------
 * catch_sigchld
 * needed to prevent defunct processes....
 *--------------------------------------------------------
 */

void catch_sigchld(int sig)
{
  int status;
  int pid;

  for (;;) {
    pid = waitpid(0,&status, WNOHANG);

    if (pid<=0) {
      break;
    }

}


/*--------------------------------------------------------
 * run_server
 * remote socket connection is on stdin/stdout
 *--------------------------------------------------------
 */

int run_server(void)
{

  struct sockaddr_in server_name;
  size_t len;
  fd_set set;

  /*
   * keep statistiscs
   */
  int request_bytes = 0; 
  int answer_bytes = 0;
#ifdef COMPRESS
  int compressed_answer_bytes = 0;
#endif

  logmessage("run_server");

  server_sock = -1;
  signal (SIGTERM, catch_signal);


  /* 
   * Connect to Illustra
   * Create a socket
   */
  client_sock = socket(PF_INET, SOCK_STREAM, 0);
  if (client_sock == -1) {
    logerror("socket:");
    return(-1);
  }



  /*
   * connect to where?
   */
  server_name.sin_family = AF_INET;
  server_name.sin_port = htons (ILLUSTRA_PORT);  
  server_name.sin_addr.s_addr = htonl (INADDR_ANY); /* means localhost */
  len = sizeof(server_name);
  
  /*
   * connect to server
   */

  if (connect(client_sock,(struct sockaddr *) &server_name, len) < 0) {
    logerror("connect");
    return(-1);
  }


  /*
   * use select to detect activity on socket ports
   */

  for(;;) {
    
    int i;

    FD_ZERO(&set);
    FD_SET(client_sock, &set);
    FD_SET(0, &set);
    
    if (select(FD_SETSIZE, &set, NULL, NULL, NULL)<0) {
      logerror("select");
      return(-1);
    }

    if (FD_ISSET(client_sock, &set)) {
      

      /*
       * Illustra is sending something, forward
       * to stdout (fd 1)
       */

      char buf[BUFSIZE];
#ifdef COMPRESS 
      unsigned char bufc[BUFSIZEC];
#endif

      int n;

      n = read(client_sock, &buf, BUFSIZE);
      if (n==0) { /* eof reached */
	close(client_sock);
#ifdef COMPRESS
	logstatistics(request_bytes, answer_bytes, compressed_answer_bytes);
#else
	logstatistics(request_bytes, answer_bytes);
#endif
	return(1);
      } else if (n == -1) {
	logerror("read");
	return(-1);
      } else {
#ifdef COMPRESS
	unsigned long lenc = BUFSIZEC;
	if (compress(bufc, &lenc, (unsigned char*) buf, (unsigned long) n)!=0) {
	  logmessage("Error in compress");
	  return(-1);
	}
	write(1, (char *)&bufc, (int)lenc);
	answer_bytes += n;
	compressed_answer_bytes += (int)lenc;
#else
	write(1, &buf, n);
	answer_bytes += n;
#endif
      }
    }

    if (FD_ISSET(0, &set)) {
      

      /*
       * There is info waiting on stdin (fd 0)
       * forward this to Illustra
       */

      char buf[BUFSIZE];
      int n;

      n = read(0, &buf, BUFSIZE);
      if (n==0) { /* eof reached */
	close(client_sock);
#ifdef COMPRESS
	logstatistics(request_bytes, answer_bytes, compressed_answer_bytes);
#else
	logstatistics(request_bytes, answer_bytes);
#endif
	return(1);
      } else if (n == -1) {
	logerror("read");
	return(-1);
      } else {
	write(client_sock, &buf, n);
	request_bytes += n;
      }
    }
	

  }
  
    

}

/*--------------------------------------------------------
 * handle_connection
 *--------------------------------------------------------
 */


int handle_connection(int fd)
{

  pid_t pid;

  if (DEBUG)
      logmessage("handle_connection");


  if ((pid = fork()) < 0) {
    logerror("fork");
    return(-1);
  }
  else if (pid != 0) { 
    /* we are the parent */
   return(0);
  } 

  /*
   * we are the child 
   * first reset stderr, stdin, stdout to fd
   * lets do this really dirty, not with this fancy
   * STDIN_FILENO etc.
   * Because my memory often fails me, here the numbers:
   *    stdin 0
   *    stdout 1
   *    stderr 2
   */
  dup2(fd,0);
  close(fd);
  dup2(0,1);
  dup2(0,2);
  

  /*
   * handle the service
   */
  run_server();

  /*
   * and _exit (important!)
   */

  logmessage("_exit");
  _exit(0);

}

/*--------------------------------------------------------
 * logmessage
 * write a log line
 * we prefix it with current date/time
 *--------------------------------------------------------
 */

int logmessage(char *s)
{
  FILE *f = NULL;
  time_t curtime; 


  f = fopen(LOGFILE,"a");
  if (f != NULL) {
    char buf[1024];

    curtime = time (NULL);
    sprintf(buf,"%s",ctime(&curtime));
    /* get rid of \n */
    if (strlen(buf)>0) 
      buf[strlen(buf)-1] = '\0';
    fprintf(f, "bridged %s  %s\n",buf, s);
    fclose(f);   
  } 

  return(0);
}

/*--------------------------------------------------------
 * logerror
 *--------------------------------------------------------
 */

int logerror(char *s)
{

  char *errmess = strerror(errno);
  char mess[1024];

  sprintf(mess,"%s: %s",s,errmess);
  logmessage(mess);

  return(0);
}

/*--------------------------------------------------------
 * logstatististics
 *--------------------------------------------------------
 */

#ifdef COMPRESS
int logstatistics(int request_bytes, int answer_bytes, int compressed_answer_bytes)
#else
int logstatistics(int request_bytes, int answer_bytes)
#endif
{

  char mess[1024];

#ifdef COMPRESS
  sprintf(mess,"Queries took %d bytes, results took %d bytes (%d uncompressed)",
	  request_bytes, compressed_answer_bytes, answer_bytes);
#else
  sprintf(mess,"Queries took %d bytes, results took %d bytes",
	  request_bytes, answer_bytes);
#endif
  logmessage(mess);

  return(0);
}



/*--------------------------------------------------------
 * catch_signal
 * when 'kill nnn' is given we want to stop, but
 * first we want to close the sockets so we can restart
 * bridged quickly (otherwise port will be in use for
 * another few minutes). 
 *--------------------------------------------------------
 */

void catch_signal(int sig)
{

  if (server_sock != -1)
    close(server_sock);
  if (client_sock != -1)
    close(client_sock);
  exit(0);
}


